/**
 * @(#) Quake2PakReader
 *
 * Reader class for the Quake 2 Pak file. Does not read in the lump data
 * and store it, but read() can be used to read an individual lump to obtain
 * that data.
 *
 * Part of blackmane.quake2.pak
 */
 
package blackmane.quake2.pak;

import blackmane.quake2.pak.*;
import blackmane.quake2.utility.*;
import java.io.*;

public class Quake2PakReader {
	// these are the data items for the reader
	protected Quake2PakHeader	header = null;
	protected Quake2PakDirectoryEntry[]	directory = null;
	
	// the number of items in the directory
	int numEntries;
	
	
	// the following lets us access the files 
	protected RandomAccessFile	file;	// need a RandomAccessFile because we will be
																	// jumping around a lot in it.
	protected boolean	open = false;	// flag to determine whether the file is open
	protected boolean valid = false; // flag to determine whether the file's info has been read in
	String pakName;	// location of the pak file.
	
	/**
	 * Pass it the name of the pak file you want to open. The file
	 * will be opened in read-only mode.
	 * 
	 * @param filename	The name of the file to open, both the path and file name.
	 * @exception Quake2PakException if the file's magic number is bad
	 *            IOException if an error occured when opening or reading from the file
	 */
	public Quake2PakReader(String filename) throws IOException, Quake2PakException {
		pakName = filename;
		file = new RandomAccessFile(pakName, "r");
		
		// now we attempt to read in the header information and
		// then the directory information, provided that the header's
		// magic number matches
		// just to be extra sure, do a seek to the start of the file,
		// even though it is totally unnecessary :)
		file.seek(0);
		// read in the header
		header = new Quake2PakHeader(IntegerUtility.intSwap(file.readInt()),
		IntegerUtility.intSwap(file.readInt()), IntegerUtility.intSwap(file.readInt()));
		
		// only do the following if we have a valid magic number
		if (header.validPak()) {
			numEntries = header.directoryLength()/64;
			directory = new Quake2PakDirectoryEntry[numEntries];
		
			int i;
			// read in the directory entries, after seeking to their
			// position in the file.
			// need to have nameBuffer in the loop otherwise the readFully will
			// not refill it from the start.
			file.seek(header.directoryOffset());
			for (i=0; i<numEntries; i++) {
				byte[] nameBuffer = new byte[56];
				file.readFully(nameBuffer);
				directory[i] = new Quake2PakDirectoryEntry(nameBuffer, IntegerUtility.intSwap(file.readInt()), 
					IntegerUtility.intSwap(file.readInt()));
			}
			// only now can we say that this file has been opened and is ready for use.
			open = true;
			valid = true;
		}
		else
			throw new Quake2PakException("Bad magic number");
	}
	
	
	/**
	 * Closes the file. Further read() accesses will fail after this has been
	 * called.
	 * 
	 */
	public void close() throws IOException {
		if (open) {
			file.close();
			open = false;
		}
	}
	
	/**
	 * Reads a lump from the file and returns a byte array containing the
	 * information.
	 * The passed integer is the directory index to use.
	 * The array being passed back will be set to null if either the
	 * file is not open or the index into the directory was not valid.
	 *
	 * @param index	the index of the lump in the directory to fetch.
	 * @return an array of bytes containing the lump
	 * @exception Quake2PakException if the file is not open or not valid
	 *             or if the passed index is not in range.
	 */
	public byte[] read(int index)	throws IOException, Quake2PakException {
		byte[] buffer = null;
		// do some checks on the index into the directory.
		if (open && index < numEntries && index > -1) {
			buffer = new byte[directory[index].length()];
			
			// seek to the position of the data.
			file.seek(directory[index].position());
			// read in the data
			file.readFully(buffer);
		}
		else
			throw new Quake2PakException("File not open or index out of range");
		
		return buffer;
	}
	
	/*
	 * Returns the header information for this pak file.
	 * 
	 * @return the Quake2PakHeader
	 * @exception Quake2PakException if the pak fiel was not opened validly.
	 */
	public Quake2PakHeader getHeader() throws Quake2PakException {
		if (valid) {
			return header;
		}

		throw new Quake2PakException("Valid flag not set - file not valid in some form");
	}
	
	/*
	 * Returns the directory information for this pak file.
	 *
	 * @return an array of Quake2PakDirectoryEntry objects.
	 * @exception QUake2PakException if the pak file was not opened validly.
	 */
	public Quake2PakDirectoryEntry[] getDirectory()  throws Quake2PakException {
		if (valid) {
			return directory;
		}

		throw new Quake2PakException("Valid flag not set - file not valid in some form");
	}
	
	/**
	 * Tell whether the pak file has had its header and directory information
	 * read in correctly.
	 *
	 * @return <code>true</code> if the pak file's header and directory information
	 *          have been read in;
	 *         <code>false</code> otherwise.
	 */
	public boolean valid() {
		return valid;
	}
	
	/**
	 * Tell whether the pak file is currently open (reads will work), or
	 * closed (reads will not work).
	 *
	 * @return <code>true</code> if the pak file can still perform read()s;
	 *         <code>false</code> otherwise.
	 */
	public boolean open() {
		return open;
	}
	
	
	/**
	 * Utility function, it returns an array of chopped strings
	 * of the lump names.
	 *
	 * @return an array of Strings that contain the internal names from
	 *         the lump directory, which have had any 0 bytes at the end removed.
	 * @exception Quake2PakException is the pak file was not validly opened.
	 */
	public String[] fileNames() throws Quake2PakException {
		String[] stringArray = null;
		int i;
		if (valid) {
			stringArray = new String[numEntries];
			for (i=0; i<numEntries; i++)
				stringArray[i] = new String(directory[i].choppedName());
			return stringArray;
		}

		throw new Quake2PakException("Valid flag not set, no name list to create");	
	}
	
}
	
	