/**
 * @(#)Quake2PakWriter
 *
 * Used to create new quake 2 pak files from scratch.
 *
 * Part of blackmane.quake2.pak;
 */

package blackmane.quake2.pak;

// this part of the package requires some of the other parts
import blackmane.quake2.pak.*;
import blackmane.quake2.utility.*; 
import java.io.*;

public class Quake2PakWriter {
	// the data items for the writer
	protected	Quake2PakHeader header;
	protected Quake2PakDirectoryEntry[] directory;
	
	// the file we are going to create
	RandomAccessFile writeFile;
	// and something with which to create the file in the first place
	File createFile;
	
	// the number of lumps in the pak file.
	int numEntries;
	
	String[] inputFileNames;
	String	pakFileName;
	
	/**
	 * This creates the header and directory structures, but doesn't
	 * actually write out the file yet - don't know why, just felt like doing it that way.
	 * This will work out the <code>header</code> and the <code>directory</code> array
	 * and have the information correct, with the directory set to be written out at the
	 * end of the pak file.
	 *
	 * @param  name the name of the pak file to create;
	 * @param  inputFiles the list of directory+path files that will be put into the pak file;
	 * @param  internalNames the list of names that the real names will appear as in the pak file
	 * @exception IOException if there is a problem opening one of the specified files in inputFiles.
	 */
	public Quake2PakWriter(String name, String[] inputFiles, String[] internalNames) throws IOException {
		File testFile;	// we need to test each file's existance
		int dataLength = 0;
		int currentOfs = 12;
		byte[] nameBuffer;
		
		numEntries = inputFiles.length;
		directory = new Quake2PakDirectoryEntry[numEntries];
		
		pakFileName = name;
		inputFileNames = inputFiles;
		
		int i;
		for (i=0; i<numEntries; i++) {
			testFile = new File(inputFiles[i]);
			// make sure we have a valid file first.
			if (testFile.isFile()) {
				// need to cast the long to an int.
				dataLength = (int)testFile.length();
				
				// convert the internal name to a quake pak name (converting \ to /)
				nameBuffer = NameToQuakeName(internalNames[i]);
				directory[i] = new Quake2PakDirectoryEntry(nameBuffer, currentOfs, dataLength);
				currentOfs += dataLength;
			}
			else
				throw new IOException(inputFiles[i] + " is not a file");
		}
		
		// create the header information
		header = new Quake2PakHeader(currentOfs, numEntries*64);
	}
	
	/**
	 * Used to convert a String into an array of bytes after
	 * converting all instances of \ into /. This is because quake
	 * uses / as a separator and java may use a \ (platform decides).
	 *
	 * @return an array of bytes (56 bytes in length) that are used for the internal name
	 * @param name the name to convert to quake2 format.
	 */
	public static byte[] NameToQuakeName(String name) {
		StringBuffer	nameBuffer = new StringBuffer(name);
		int index;		
		for (index = 0 ; index < nameBuffer.length() ; index++) {
			// look for all File.separators and replace them with /'s
			if (nameBuffer.charAt(index) == File.separatorChar) {
				nameBuffer.setCharAt(index, '/');
			}
		}
		
		// this will expand the buffer if it is less than 56 characters
		// in length or truncate it to a maximum of 56 characters - it is
		// up to the hsot program to make sure that the entered name is never greater
		// than 56 characters
		nameBuffer.setLength(56);
		
		byte[] byteArray = nameBuffer.toString().getBytes();
		return byteArray;
	}
	
	/**
	 * Creates the pak file from scratch. This will overwrite any existing file
	 * of the same name, and will create any parent directories necessary.
	 *
	 * @exception IOException if an error occurs accessing a file or reading a file
	 */
	public void create() throws IOException {
		// make any necessary directories for this new file
		String paths = null;
		createFile = new File(pakFileName);
		paths = createFile.getParent();
		// we need to check whether this is null or not because getParent() will
		// return null if there were no File.separator() characters. If there were
		// some, then we have to create the directories for this new file.
		if (paths != null)
			createFile.mkdirs();
		
		// we use a RandomAccessFile because it has both writeInt and write(byte[])
		// for us to use.
		writeFile = new RandomAccessFile(createFile, "rw");
		File inputFile;
		FileInputStream inputStream;
		int i;
		byte[] buffer;
		
		// first we write out the header information, remembering 
		// to use the intSwap routine all the time.
		writeFile.writeInt(IntegerUtility.intSwap(header.magic()));
		writeFile.writeInt(IntegerUtility.intSwap(header.directoryOffset()));
		writeFile.writeInt(IntegerUtility.intSwap(header.directoryLength()));
		
		// now write out the lumps
		for (i=0; i<numEntries; i++) {
			inputFile = new File(inputFileNames[i]);
			inputStream = new FileInputStream(inputFile);
			// need to cast the long to an int
			buffer = new byte[(int)inputFile.length()];
			inputStream.read(buffer);
			
			writeFile.write(buffer);
			inputStream.close();
	}
		
		// now write out the directory
		for (i=0; i<numEntries; i++) {
			writeFile.write(directory[i].name());
			writeFile.writeInt(IntegerUtility.intSwap(directory[i].position()));
			writeFile.writeInt(IntegerUtility.intSwap(directory[i].length()));
		}
		
		writeFile.close();

	}
	
}